package main

import (
	"errors"
	"flag"
	"fmt"
	"os"
	"os/exec"
	"path/filepath"
	"runtime"
	"strconv"
	"strings"
	"sync"
	"time"

	"git.milar.in/milarin/channel"
)

// globals
var (
	ModulePath  string
	ProjectName string
	Runner      channel.Runner
)

// command line arguments
var (
	OS   = flag.String("os", "", "comma-separated list of operating systems to compile for (empty for all)")
	Arch = flag.String("arch", "", "comma-separated list of architectures to compile for (empty for all)")

	Silent     = flag.Bool("s", false, "silent mode (no output)")
	NoCompress = flag.Bool("c", false, "dont compress any executables")
	NumThreads = flag.Int("t", runtime.NumCPU(), "amount of threads (0 = infinite)")
)

func main() {
	var err error
	flag.Parse()

	ModulePath, err = filepath.Abs(flag.Arg(0))
	if err != nil {
		panic(err)
	}

	if _, err := exec.LookPath("go"); err != nil {
		ColorError.Fprintln(os.Stderr, "go not found in PATH. compilation not possible")
		return
	}

	if !*NoCompress {
		if _, err := exec.LookPath("upx"); err != nil {
			ColorWarn.Fprintln(os.Stderr, "upx not found in PATH. file compression not possible")
			*NoCompress = true
		}
	}

	if err := FillCompileConfigs(); err != nil {
		ColorError.Fprintln(os.Stderr, fmt.Errorf("target architectures could not be determined: %w", err))
		return
	}

	if _, err := os.Stat(filepath.Join(ModulePath, "go.mod")); errors.Is(err, os.ErrNotExist) {
		ColorError.Fprintf(os.Stderr, "no Go module found at '%s'\n", ModulePath)
		return
	}

	if err := DetermineProjectName(); err != nil {
		ColorError.Fprintln(os.Stderr, fmt.Errorf("project name could not be determined: %w", err))
		return
	}

	ch := make(chan *CompileReport, len(CompileConfigs))
	wg := new(sync.WaitGroup)

	doneCh := make(chan struct{})
	go Watch(ch, doneCh)

	Runner = GetRunner()

	start := time.Now()

	for _, cfg := range CompileConfigs {
		cfg := cfg
		wg.Add(1)
		go Runner.Run(func() { Compile(cfg, ch, wg) })
	}

	wg.Wait()
	close(ch)

	end := time.Now()

	<-doneCh // wait for Watch

	if !*Silent {
		fmt.Printf("compilation took %s (using %s threads)\n", end.Sub(start), GetThreadCountString())
	}
}

func DetermineProjectName() error {
	data, err := os.ReadFile(filepath.Join(ModulePath, "go.mod"))
	if err != nil {
		return err
	}

	for _, line := range strings.Split(string(data), "\n") {
		if strings.HasPrefix(line, "module ") {
			fullName := strings.TrimPrefix(line, "module ")
			parts := strings.Split(fullName, "/")
			ProjectName = parts[len(parts)-1]
			return nil
		}
	}

	return errors.New("malformed syntax in go.mod file")
}

func GetRunner() channel.Runner {
	if *NumThreads <= 0 {
		return channel.NewUnlimitedRunner()
	} else {
		return channel.NewLimitedRunner(*NumThreads)
	}
}

func GetThreadCountString() string {
	if *NumThreads == 0 {
		return "infinitely many"
	}
	return strconv.Itoa(*NumThreads)
}
